@@ -1,6 +1,6 @@
 ;;; tramp-cmds.el --- Interactive commands for Tramp  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2007-2019 Free Software Foundation, Inc.
+;; Copyright (C) 2007-2025 Free Software Foundation, Inc.
 
 ;; Author: Michael Albinus <michael.albinus@gmx.de>
 ;; Keywords: comm, processes
@@ -31,11 +31,17 @@
 (require 'tramp)
 
 ;; Pacify byte-compiler.
+(declare-function dired-advertise "dired")
+(declare-function dired-get-file-for-visit "dired")
+(declare-function dired-unadvertise "dired")
 (declare-function mml-mode "mml")
 (declare-function mml-insert-empty-tag "mml")
 (declare-function reporter-dump-variable "reporter")
+(defvar mm-7bit-chars)
 (defvar reporter-eval-buffer)
 (defvar reporter-prompt-for-summary-p)
+(defvar tramp-repository-branch)
+(defvar tramp-repository-version)
 
 ;;;###tramp-autoload
 (defun tramp-change-syntax (&optional syntax)
@@ -46,27 +52,86 @@ SYNTAX can be one of the symbols `default' (default),
    (let ((input (completing-read
 		 "Enter Tramp syntax: " (tramp-syntax-values) nil t
 		 (symbol-name tramp-syntax))))
-     (unless (string-equal input "")
+     (unless (string-empty-p input)
        (list (intern input)))))
   (when syntax
     (customize-set-variable 'tramp-syntax syntax)))
 
+;;;###tramp-autoload
+(defun tramp-enable-method (method)
+  "Enable optional METHOD if possible."
+  (interactive
+   (list
+    (completing-read
+     "method: "
+     (tramp-compat-seq-keep
+      (lambda (x)
+	(when-let* ((name (symbol-name x))
+		    ;; It must match `tramp-enable-METHOD-method'.
+		    ((string-match
+		      (rx "tramp-enable-"
+			  (group (regexp tramp-method-regexp))
+			  "-method")
+		      name))
+		    (method (match-string 1 name))
+		    ;; It must not be enabled yet.
+		    ((not (assoc method tramp-methods))))
+	  method))
+      ;; All method enabling functions.
+      (mapcar
+       #'intern (all-completions "tramp-enable-" obarray #'functionp))))))
+
+  (when-let* (((not (assoc method tramp-methods)))
+	      (fn (intern (format "tramp-enable-%s-method" method)))
+	      ((functionp fn)))
+    (funcall fn)
+    (message "Tramp method \"%s\" enabled" method)))
+
+;; Use `match-buffers' starting with Emacs 29.1.
+;;;###tramp-autoload
 (defun tramp-list-tramp-buffers ()
   "Return a list of all Tramp connection buffers."
   (append
    (all-completions
     "*tramp" (mapcar #'list (mapcar #'buffer-name (buffer-list))))
    (all-completions
-    "*debug tramp" (mapcar #'list (mapcar #'buffer-name (buffer-list))))))
+    "*debug tramp" (mapcar #'list (mapcar #'buffer-name (buffer-list))))
+   (all-completions
+    "*trace tramp" (mapcar #'list (mapcar #'buffer-name (buffer-list))))))
 
+;; Use `match-buffers' starting with Emacs 29.1.
+;;;###tramp-autoload
 (defun tramp-list-remote-buffers ()
-  "Return a list of all buffers with remote `default-directory'."
-  (delq
-   nil
-   (mapcar
-    (lambda (x)
-      (with-current-buffer x (when (tramp-tramp-file-p default-directory) x)))
-    (buffer-list))))
+  "Return a list of remote buffers, excluding internal Tramp buffers.
+A buffer is considered remote if either its `default-directory' or
+`buffer-file-name' is a remote file name."
+  (tramp-compat-seq-keep
+   (lambda (buffer)
+     (when (tramp-tramp-file-p
+            (or (buffer-file-name buffer)
+                (tramp-get-default-directory buffer)))
+       buffer))
+   (buffer-list)))
+
+;;;###tramp-autoload
+(defun tramp-list-remote-buffer-connections ()
+  "Return a list of all remote buffer connections.
+A buffer is considered remote if either its `default-directory' or
+`buffer-file-name' is a remote file name."
+  (seq-uniq
+   (mapcar (lambda (buffer)
+             (or
+              (when (buffer-file-name buffer)
+                (file-remote-p (buffer-file-name buffer)))
+              (when (tramp-get-default-directory buffer)
+                (file-remote-p (tramp-get-default-directory buffer)))))
+           ;; Eliminate false positives from internal Tramp buffers.
+           (seq-remove
+            (lambda (buffer)
+              (member (buffer-name buffer) (tramp-list-tramp-buffers)))
+            (tramp-list-remote-buffers)))))
+
+;;; Cleanup
 
 ;;;###tramp-autoload
 (defvar tramp-cleanup-connection-hook nil
@@ -74,12 +139,15 @@ SYNTAX can be one of the symbols `default' (default),
 Each function is called with the current vector as argument.")
 
 ;;;###tramp-autoload
-(defun tramp-cleanup-connection (vec &optional keep-debug keep-password)
+(defun tramp-cleanup-connection
+    (vec &optional keep-debug keep-password keep-processes)
   "Flush all connection related objects.
-This includes password cache, file cache, connection cache,
-buffers.  KEEP-DEBUG non-nil preserves the debug buffer.
-KEEP-PASSWORD non-nil preserves the password cache.
-When called interactively, a Tramp connection has to be selected."
+This includes password cache, file cache, connection cache, buffers,
+processes.  KEEP-DEBUG non-nil preserves the debug and trace buffer.
+KEEP-PASSWORD non-nil preserves the password cache.  KEEP-PROCESSES
+non-nil preserves the asynchronous processes.  When called
+interactively, a Tramp connection has to be selected."
+  (declare (completion tramp-active-command-completion-p))
   (interactive
    ;; When interactive, select the Tramp remote identification.
    ;; Return nil when there is no Tramp connection.
@@ -107,35 +175,44 @@ When called interactively, a Tramp connection has to be selected."
     ;; suppressed.
     (setq tramp-current-connection nil)
 
-    ;; Flush file cache.
-    (tramp-flush-directory-properties vec "")
-
-    ;; Flush connection cache.
-    (when (processp (tramp-get-connection-process vec))
-      (tramp-flush-connection-properties (tramp-get-connection-process vec))
-      (delete-process (tramp-get-connection-process vec)))
-    (tramp-flush-connection-properties vec)
-
     ;; Cancel timer.
     (dolist (timer timer-list)
-      (when (and (eq (timer--function timer) 'tramp-timeout-session)
+      (when (and (eq (timer--function timer) #'tramp-timeout-session)
 		 (tramp-file-name-equal-p vec (car (timer--args timer))))
 	(cancel-timer timer)))
 
+    ;; Delete processes.
+    (dolist (key (hash-table-keys tramp-cache-data))
+      (when (and (processp key)
+		 (tramp-file-name-equal-p (process-get key 'tramp-vector) vec)
+		 (or (not keep-processes)
+		     (eq key (tramp-get-process vec))))
+	(tramp-flush-connection-properties key)
+	(ignore-errors (delete-process key))))
+
     ;; Remove buffers.
     (dolist
 	(buf (list (get-buffer (tramp-buffer-name vec))
 		   (unless keep-debug
 		     (get-buffer (tramp-debug-buffer-name vec)))
-		   (tramp-get-connection-property vec "process-buffer" nil)))
+		   (unless keep-debug
+		     (get-buffer (tramp-trace-buffer-name vec)))
+		   (tramp-get-connection-property vec " connected")))
       (when (bufferp buf) (kill-buffer buf)))
 
+    ;; Flush file cache.
+    (tramp-flush-directory-properties vec "/")
+
+    ;; Flush connection cache.
+    (tramp-flush-connection-properties vec)
+
     ;; The end.
     (run-hook-with-args 'tramp-cleanup-connection-hook vec)))
 
 ;;;###tramp-autoload
 (defun tramp-cleanup-this-connection ()
   "Flush all connection related objects of the current buffer's connection."
+  (declare (completion tramp-command-completion-p))
   (interactive)
   (and (tramp-tramp-file-p default-directory)
        (tramp-cleanup-connection
@@ -149,17 +226,19 @@ When called interactively, a Tramp connection has to be selected."
 (defun tramp-cleanup-all-connections ()
   "Flush all Tramp internal objects.
 This includes password cache, file cache, connection cache, buffers."
+  (declare (completion tramp-active-command-completion-p))
   (interactive)
 
-  ;; Unlock Tramp.
-  (setq tramp-locked nil)
-
   ;; Flush password cache.
   (password-reset)
 
   ;; Flush file and connection cache.
   (clrhash tramp-cache-data)
 
+  ;; Initialize the cache version.
+  (tramp-set-connection-property
+   tramp-cache-version "tramp-version" tramp-version)
+
   ;; Remove ad-hoc proxies.
   (let ((proxies tramp-default-proxies-alist))
     (while proxies
@@ -176,26 +255,115 @@ This includes password cache, file cache, connection cache, buffers."
   ;; Cancel timers.
   (cancel-function-timers 'tramp-timeout-session)
 
-  ;; Remove buffers.
+  ;; Remove processes and buffers.
   (dolist (name (tramp-list-tramp-buffers))
+    (when (processp (get-buffer-process name)) (delete-process name))
     (when (bufferp (get-buffer name)) (kill-buffer name)))
 
   ;; The end.
   (run-hooks 'tramp-cleanup-all-connections-hook))
 
+(defcustom tramp-cleanup-some-buffers-hook nil
+  "Hook for `tramp-cleanup-some-buffers'.
+The functions determine which buffers shall be killed.  This
+happens when at least one of the functions returns non-nil.  The
+functions are called with `current-buffer' set."
+  :group 'tramp
+  :version "30.1"
+  :type 'hook
+  :link '(info-link :tag "Tramp manual" "(tramp) Cleanup remote connections"))
+
+(add-hook 'tramp-cleanup-some-buffers-hook
+	  #'buffer-file-name)
+
+(defun tramp-dired-buffer-p ()
+  "Return t if current buffer runs `dired-mode'."
+  (declare (tramp-suppress-trace t))
+  (derived-mode-p 'dired-mode))
+
+(add-hook 'tramp-cleanup-some-buffers-hook
+	  #'tramp-dired-buffer-p)
+
+(defvar tramp-tainted-remote-process-buffers nil
+  "List of process buffers to be cleaned up.")
+
+(defun tramp-delete-tainted-remote-process-buffer-function ()
+  "Delete current buffer from `tramp-tainted-remote-process-buffers'."
+  (declare (tramp-suppress-trace t))
+  (setq tramp-tainted-remote-process-buffers
+	(delete (current-buffer) tramp-tainted-remote-process-buffers)))
+
 ;;;###tramp-autoload
-(defun tramp-cleanup-all-buffers ()
-  "Kill all remote buffers."
+(defun tramp-taint-remote-process-buffer (buffer)
+  "Mark buffer as related to remote processes."
+  ;; (declare (tramp-suppress-trace t))
+  (add-to-list 'tramp-tainted-remote-process-buffers buffer))
+
+;; We cannot use the `declare' form for `tramp-suppress-trace' in
+;; autoloaded functions, because the tramp-loaddefs.el generation
+;; would fail.
+(function-put #'tramp-taint-remote-process-buffer 'tramp-suppress-trace t)
+
+(add-hook 'kill-buffer-hook
+	  #'tramp-delete-tainted-remote-process-buffer-function)
+(add-hook 'tramp-unload-hook
+	  (lambda ()
+	    (remove-hook 'kill-buffer-hook
+			 #'tramp-delete-tainted-remote-process-buffer-function)))
+
+(defun tramp-remote-process-p ()
+  "Return t if current buffer belongs to a remote process."
+  (memq (current-buffer) tramp-tainted-remote-process-buffers))
+
+(add-hook 'tramp-cleanup-some-buffers-hook
+	  #'tramp-remote-process-p)
+
+;;;###tramp-autoload
+(defun tramp-cleanup-some-buffers ()
+  "Kill some remote buffers.
+A buffer is killed when it has a remote `default-directory', and
+one of the functions in `tramp-cleanup-some-buffers-hook' returns
+non-nil."
+  (declare (completion tramp-active-command-completion-p))
   (interactive)
 
   ;; Remove all Tramp related connections.
   (tramp-cleanup-all-connections)
 
-  ;; Remove all buffers with a remote default-directory.
+  ;; Remove all buffers with a remote default-directory which fit the hook.
   (dolist (name (tramp-list-remote-buffers))
-    (when (bufferp (get-buffer name)) (kill-buffer name))))
+    (and (buffer-live-p (get-buffer name))
+	 (with-current-buffer name
+	   (run-hook-with-args-until-success 'tramp-cleanup-some-buffers-hook))
+	 (kill-buffer name))))
 
 ;;;###tramp-autoload
+(defun tramp-cleanup-all-buffers ()
+  "Kill all remote buffers."
+  (declare (completion tramp-active-command-completion-p))
+  (interactive)
+  (let ((tramp-cleanup-some-buffers-hook '(always)))
+    (tramp-cleanup-some-buffers)))
+
+;;;###tramp-autoload
+(defun tramp-cleanup-bufferless-connections ()
+  "Flush connection-related objects for which no buffer exists.
+A bufferless connection is one for which no live buffer's
+`buffer-file-name' or `default-directory' is associated with that
+connection, except for Tramp internal buffers.
+Display a message of cleaned-up connections."
+  (interactive)
+  (when-let* ((bufferless-connections
+               (seq-difference
+                (mapcar #'tramp-make-tramp-file-name (tramp-list-connections))
+                (tramp-list-remote-buffer-connections))))
+    (message "Cleaning up %s" (string-join bufferless-connections ", "))
+    (dolist (connection bufferless-connections)
+      (tramp-cleanup-connection
+       (tramp-dissect-file-name connection 'noexpand)))))
+
+;;; Rename
+
 (defcustom tramp-default-rename-alist nil
   "Default target for renaming remote buffer file names.
 This is an alist of cons cells (SOURCE . TARGET).  The first
@@ -216,14 +384,15 @@ expression which always matches."
   :group 'tramp
   :version "27.1"
   :type '(repeat (cons (choice :tag "Source regexp" regexp sexp)
-		       (choice :tag "Target   name" string (const nil)))))
+		       (choice :tag "Target   name" string (const nil))))
+  :link '(info-link :tag "Tramp manual" "(tramp) Renaming remote files"))
 
-;;;###tramp-autoload
 (defcustom tramp-confirm-rename-file-names t
   "Whether renaming a buffer file name must be confirmed."
   :group 'tramp
   :version "27.1"
-  :type 'boolean)
+  :type 'boolean
+  :link '(info-link :tag "Tramp manual" "(tramp) Renaming remote files"))
 
 (defun tramp-default-rename-file (string)
   "Determine default file name for renaming according to STRING.
@@ -237,10 +406,10 @@ function returns nil"
 	  (host (or (file-remote-p string 'host) ""))
 	  item result)
       (while (setq item (pop tdra))
-	(when (string-match-p (or (eval (car item)) "") string)
+	(when (string-match-p (or (eval (car item) t) "") string)
 	  (setq tdra nil
 		result
-		(format-spec
+		(tramp-format-spec
 		 (cdr item) (format-spec-make ?m method ?u user ?h host)))))
       result)))
 
@@ -287,6 +456,7 @@ without confirmation if the prefix argument is non-nil.
 
 The remote connection identified by SOURCE is flushed by
 `tramp-cleanup-connection'."
+  (declare (completion tramp-active-command-completion-p))
   (interactive
    (let ((connections
 	  (mapcar #'tramp-make-tramp-file-name (tramp-list-connections)))
@@ -297,9 +467,9 @@ The remote connection identified by SOURCE is flushed by
 	 (read-file-name-function #'read-file-name-default)
 	  source target)
      (if (null connections)
-	 (tramp-user-error nil "There are no remote connections.")
+	 (tramp-user-error nil "There are no remote connections")
        (setq source
-	     ;; Likely, the source remote connection is broken. So we
+	     ;; Likely, the source remote connection is broken.  So we
 	     ;; shall avoid any action on it.
 	     (let (non-essential)
 	       (completing-read-default
@@ -337,7 +507,7 @@ The remote connection identified by SOURCE is flushed by
 		      (dir (tramp-rename-read-file-name-dir default))
 		      (init (tramp-rename-read-file-name-init default))
 		      (tramp-ignored-file-name-regexp
-		       (regexp-quote (file-remote-p source))))
+		       (rx (literal (file-remote-p source)))))
 		 (read-file-name-default
 		  "Enter new Tramp connection: "
 		  dir default 'confirm init #'file-directory-p)))))
@@ -345,21 +515,20 @@ The remote connection identified by SOURCE is flushed by
      (list source target)))
 
   (unless (tramp-tramp-file-p source)
-    (tramp-user-error nil "Source %s must be remote." source))
+    (tramp-user-error nil "Source %s must be remote" source))
   (when (null target)
     (or (setq target (tramp-default-rename-file source))
 	(tramp-user-error
 	 nil
-	 (eval-when-compile
-	   (concat "There is no target specified.  "
-		   "Check `tramp-default-rename-alist' for a proper entry.")))))
+	 (concat "There is no target specified.  "
+		 "Check `tramp-default-rename-alist' for a proper entry"))))
   (when (tramp-equal-remote source target)
-    (tramp-user-error nil "Source and target must have different remote."))
+    (tramp-user-error nil "Source and target must have different remote"))
 
   ;; Append local file name if none is specified.
   (when (string-equal (file-remote-p target) target)
-    (setq target (concat target (file-remote-p source 'localname))))
-  ;; Make them directoy names.
+    (setq target (concat target (tramp-file-local-name source))))
+  ;; Make them directory names.
   (setq source (directory-file-name source)
 	target (directory-file-name target))
 
@@ -378,12 +547,11 @@ ESC or `q' to quit without changing further buffers,
 	(dolist (buffer (tramp-list-remote-buffers))
           (switch-to-buffer buffer)
 	  (let* ((bfn (buffer-file-name))
-		 (new-bfn (and (stringp bfn)
-			       (replace-regexp-in-string
-				(regexp-quote source) target bfn)))
+		 (new-bfn (and (stringp bfn) (string-replace source target bfn)))
 		 (prompt (format-message
 			  "Set visited file name to `%s' [Type yn!eq or %s] "
-			  new-bfn (key-description (vector help-char)))))
+                          new-bfn (if (fboundp 'help-key) (help-key) ; 29.1
+                                    (key-description (vector help-char))))))
 	    (when (and (buffer-live-p buffer) (stringp bfn)
 		       (string-prefix-p source bfn)
 		       ;; Skip, and don't ask again.
@@ -427,6 +595,7 @@ Interactively, TARGET is selected from `tramp-default-rename-alist'
 without confirmation if the prefix argument is non-nil.
 
 For details, see `tramp-rename-files'."
+  (declare (completion tramp-command-completion-p))
   (interactive
    (let ((source default-directory)
 	 target
@@ -440,7 +609,7 @@ For details, see `tramp-rename-files'."
 	  nil
 	  (substitute-command-keys
 	   (concat "Current buffer is not remote.  "
-		   "Consider `\\[tramp-rename-files]' instead.")))
+		   "Consider `\\[tramp-rename-files]' instead")))
        (setq target
 	     (when (null current-prefix-arg)
 	       ;; The source remote connection shall not trigger any action.
@@ -449,7 +618,7 @@ For details, see `tramp-rename-files'."
 		      (dir (tramp-rename-read-file-name-dir default))
 		      (init (tramp-rename-read-file-name-init default))
 		      (tramp-ignored-file-name-regexp
-		       (regexp-quote (file-remote-p source))))
+		       (rx (literal (file-remote-p source)))))
 		 (read-file-name-default
 		  (format "Change Tramp connection `%s': " source)
 		  dir default 'confirm init #'file-directory-p)))))
@@ -457,11 +626,178 @@ For details, see `tramp-rename-files'."
 
   (tramp-rename-files default-directory target))
 
+;;; Run as sudo
+
+(defcustom tramp-file-name-with-method "sudo"
+  "Which method to be used in `tramp-file-name-with-sudo'."
+  :group 'tramp
+  :version "30.1"
+  :type '(choice (const "su")
+		 (const "sudo")
+		 (const "doas")
+		 (const "run0")
+		 (const "ksu"))
+  :link '(tramp-info-link :tag "Tramp manual" tramp-file-name-with-method))
+
+(defun tramp-get-file-name-with-method ()
+  "Return connection-local value of `tramp-file-name-with-method'."
+  (tramp-compat-connection-local-value tramp-file-name-with-method))
+
+(defmacro with-tramp-file-name-with-method (&rest body)
+  "Ask user for `tramp-file-name-with-method' if needed.
+Run BODY."
+  (declare (indent 0) (debug t))
+  `(let ((tramp-file-name-with-method
+          (if current-prefix-arg
+	      (completing-read
+	       "Tramp method: "
+               (mapcar
+		#'cadr (cdr (get 'tramp-file-name-with-method 'custom-type)))
+               nil t (tramp-get-file-name-with-method))
+            (tramp-get-file-name-with-method))))
+     ,@body))
+
+(defun tramp-file-name-with-sudo (filename)
+  "Convert FILENAME into a multi-hop file name with \"sudo\".
+An alternative method could be chosen with `tramp-file-name-with-method'."
+  (setq filename (expand-file-name filename))
+  (let ((default-method (tramp-get-file-name-with-method)))
+    (if (tramp-tramp-file-p filename)
+	(with-parsed-tramp-file-name filename nil
+	  (cond
+	   ;; Remote file with proper method.
+	   ((string-equal method default-method)
+	    filename)
+	   ;; Remote file on the local host.
+	   ((and
+	     (stringp tramp-local-host-regexp) (stringp host)
+	     (string-match-p tramp-local-host-regexp host))
+	    (tramp-make-tramp-file-name
+	     (make-tramp-file-name
+	      :method default-method :localname localname)))
+	   ;; Remote file with multi-hop capable method.
+	   ((tramp-multi-hop-p v)
+	    (tramp-make-tramp-file-name
+	     (make-tramp-file-name
+	      :method (tramp-find-method default-method nil host)
+	      :user (tramp-find-user default-method nil host)
+	      :host (tramp-find-host default-method nil host)
+	      :localname localname :hop (tramp-make-tramp-hop-name v))))
+	   ;; Other remote file.
+	   (t (tramp-user-error v "Multi-hop with `%s' not applicable" method))))
+      ;; Local file.
+      (tramp-make-tramp-file-name
+       (make-tramp-file-name :method default-method :localname filename)))))
+
+;; FIXME: We would like to rename this for Emacs 31.1 to a name that
+;; does not encode the default method.  It is intended as a generic
+;; privilege-elevation command.  Some ideas from bug#76974:
+;; `tramp-revert-buffer-obtain-root',
+;; `tramp-revert-buffer-as-superuser'.
+
+;;;###autoload
+(defun tramp-revert-buffer-with-sudo ()
+  "Visit the current file again with superuser, or root, permissions.
+
+By default this is done using the \"sudo\" Tramp method.
+You can customize `tramp-file-name-with-method' to change this.
+
+Interactively, with a prefix argument, prompt for a different method.
+
+If the buffer visits a file, the file is replaced.
+If the buffer runs `dired', the buffer is reverted."
+  (interactive)
+  (with-tramp-file-name-with-method
+    (cond
+     ((buffer-file-name)
+      (let ((pos (point)))
+        (find-alternate-file (tramp-file-name-with-sudo (buffer-file-name)))
+        (goto-char pos)))
+     ((tramp-dired-buffer-p)
+      (dired-unadvertise (expand-file-name default-directory))
+      (setq default-directory (tramp-file-name-with-sudo default-directory)
+	    list-buffers-directory
+	    (tramp-file-name-with-sudo list-buffers-directory))
+      (if (consp dired-directory)
+	  (setcar
+	   dired-directory (tramp-file-name-with-sudo (car dired-directory)))
+        (setq dired-directory (tramp-file-name-with-sudo dired-directory)))
+      (dired-advertise)
+      (revert-buffer)))))
+
+;; This function takes action, when `read-extended-command-predicate'
+;; is set to `command-completion-default-include-p'.
+(defun tramp-dired-buffer-command-completion-p (_symbol buffer)
+  "A predicate for Tramp interactive commands.
+They are completed by `M-x TAB' only in Dired buffers."
+  (declare (tramp-suppress-trace t))
+  (with-current-buffer buffer
+    (tramp-dired-buffer-p)))
+
+;; FIXME: See FIXME above about renaming this before Emacs 31.1.
+
+;;;###autoload
+(defun tramp-dired-find-file-with-sudo ()
+  "Visit the file or directory named on this line as the superuser.
+
+By default this is done using the \"sudo\" Tramp method.
+You can customize `tramp-file-name-with-method' to change this.
+
+Interactively, with a prefix argument, prompt for a different method."
+  ;; (declare (completion tramp-dired-buffer-command-completion-p))
+  (interactive)
+  (with-tramp-file-name-with-method
+    (find-file (tramp-file-name-with-sudo (dired-get-file-for-visit)))))
+
+;; `tramp-dired-buffer-command-completion-p' is not autoloaded, and this
+;; setting isn't either.
+(function-put
+ #'tramp-dired-find-file-with-sudo 'completion-predicate
+ #'tramp-dired-buffer-command-completion-p)
+
+;;; Recompile on ELPA
+
+;; This function takes action, when `read-extended-command-predicate'
+;; is set to `command-completion-default-include-p'.
+;;;###tramp-autoload
+(defun tramp-recompile-elpa-command-completion-p (_symbol _buffer)
+  "A predicate for `tramp-recompile-elpa'.
+It is completed by `M-x TAB' only if package.el is loaded, and
+Tramp is an installed ELPA package."
+  ;; We cannot apply `package-installed-p', this would also return the
+  ;; builtin package.
+  (and (assq 'tramp (bound-and-true-p package-alist))
+       (tramp-compat-funcall 'package--user-installed-p 'tramp)))
+
+;;;###tramp-autoload
+(defun tramp-recompile-elpa ()
+  "Recompile the installed Tramp ELPA package.
+This is needed if there are compatibility problems."
+  (declare (completion tramp-recompile-elpa-command-completion-p))
+  (interactive)
+  ;; We expect just one Tramp package is installed.
+  (when-let*
+      ((dir (tramp-compat-funcall
+	     'package-desc-dir
+	     (car (alist-get 'tramp (bound-and-true-p package-alist))))))
+    (dolist (elc (directory-files dir 'full (rx ".elc" eos)))
+      (delete-file elc))
+    (with-current-buffer (get-buffer-create byte-compile-log-buffer)
+      (let ((inhibit-read-only t))
+	(compilation-mode)
+	(goto-char (point-max))
+	(insert "\f\n")
+	(call-process
+	 (expand-file-name invocation-name invocation-directory) nil t t
+	 "-Q" "-batch" "-L" dir
+	 "--eval" (format "(byte-recompile-directory %S 0 t)" dir))
+	(message "Package `tramp' recompiled.")))))
+
 ;; Tramp version is useful in a number of situations.
 
 ;;;###tramp-autoload
 (defun tramp-version (arg)
-  "Print version number of tramp.el in minibuffer or current buffer."
+  "Print version number of tramp.el in echo area or current buffer."
   (interactive "P")
   (if arg (insert tramp-version) (message tramp-version)))
 
@@ -474,32 +810,31 @@ For details, see `tramp-rename-files'."
 (defun tramp-bug ()
   "Submit a bug report to the Tramp developers."
   (interactive)
-  (let ((reporter-prompt-for-summary-p t)
-	;; In rare cases, it could contain the password.  So we make it nil.
-	tramp-password-save-function)
+  (let ((reporter-prompt-for-summary-p t))
     (reporter-submit-bug-report
      tramp-bug-report-address	  ; to-address
      (format "tramp (%s %s/%s)" ; package name and version
 	     tramp-version tramp-repository-branch tramp-repository-version)
      (sort
-      (delq nil (mapcar
-	(lambda (x)
-	  (and x (boundp x) (cons x 'tramp-reporter-dump-variable)))
-	(append
-	 (mapcar #'intern (all-completions "tramp-" obarray #'boundp))
-	 ;; Non-tramp variables of interest.
-	 '(shell-prompt-pattern
-	   backup-by-copying
-	   backup-by-copying-when-linked
-	   backup-by-copying-when-mismatch
-	   backup-by-copying-when-privileged-mismatch
-	   backup-directory-alist
-	   password-cache
-	   password-cache-expiry
-	   remote-file-name-inhibit-cache
-	   connection-local-profile-alist
-	   connection-local-criteria-alist
-	   file-name-handler-alist))))
+      (tramp-compat-seq-keep
+       (lambda (x)
+	 (and x (boundp x) (not (get x 'tramp-suppress-trace))
+	      (cons x 'tramp-reporter-dump-variable)))
+       (append
+	(mapcar #'intern (all-completions "tramp-" obarray #'boundp))
+	;; Non-Tramp variables of interest.
+	'(shell-prompt-pattern
+	  backup-by-copying
+	  backup-by-copying-when-linked
+	  backup-by-copying-when-mismatch
+	  backup-by-copying-when-privileged-mismatch
+	  backup-directory-alist
+	  password-cache
+	  password-cache-expiry
+	  remote-file-name-inhibit-cache
+	  connection-local-profile-alist
+	  connection-local-criteria-alist
+	  file-name-handler-alist)))
       (lambda (x y) (string< (symbol-name (car x)) (symbol-name (car y)))))
 
      'tramp-load-report-modules	; pre-hook
@@ -532,9 +867,8 @@ buffer in your bug report.
 
 (defun tramp-reporter-dump-variable (varsym mailbuf)
   "Pretty-print the value of the variable in symbol VARSYM."
-  (let* ((reporter-eval-buffer (symbol-value 'reporter-eval-buffer))
-	 (val (with-current-buffer reporter-eval-buffer
-		(symbol-value varsym))))
+  (when-let* ((reporter-eval-buffer reporter-eval-buffer)
+	      (val (buffer-local-value varsym reporter-eval-buffer)))
 
     (if (hash-table-p val)
 	;; Pretty print the cache.
@@ -542,30 +876,31 @@ buffer in your bug report.
       ;; There are non-7bit characters to be masked.
       (when (and (stringp val)
 		 (string-match-p
-		  (concat "[^" (bound-and-true-p mm-7bit-chars) "]") val))
+		  (rx-to-string `(not (any ,mm-7bit-chars))) val))
 	(with-current-buffer reporter-eval-buffer
-	  (set
-	   varsym
-	   (format
-	    "(decode-coding-string (base64-decode-string \"%s\") 'raw-text)"
-	    (base64-encode-string (encode-coding-string val 'raw-text)))))))
+	  (set varsym
+	       `(decode-coding-string
+		 (base64-decode-string
+		  ,(base64-encode-string (encode-coding-string val 'raw-text)))
+		 'raw-text)))))
 
     ;; Dump variable.
-    (reporter-dump-variable varsym mailbuf)
+    (goto-char (point-max))
+    (save-excursion
+      (reporter-dump-variable varsym mailbuf))
 
     (unless (hash-table-p val)
       ;; Remove string quotation.
-      (forward-line -1)
       (when (looking-at
-	     (eval-when-compile
-	       (concat "\\(^.*\\)" "\""                       ;; \1 "
-		       "\\((base64-decode-string \\)" "\\\\"  ;; \2 \
-		       "\\(\".*\\)" "\\\\"                    ;; \3 \
-		       "\\(\")\\)" "\"$")))                   ;; \4 "
+	     (rx
+	      bol (group (* anychar)) "\""          ;; \1 "
+	      (group "(base64-decode-string ") "\\" ;; \2 \
+	      (group "\"" (* anychar)) "\\"         ;; \3 \
+	      (group "\")") "\"" eol))              ;; \4 "
 	(replace-match "\\1\\2\\3\\4")
 	(beginning-of-line)
-	(insert " ;; Variable encoded due to non-printable characters.\n"))
-      (forward-line 1))
+	(insert " ;; Variable encoded due to non-printable characters.\n")))
+    (goto-char (point-max))
 
     ;; Reset VARSYM to old value.
     (with-current-buffer reporter-eval-buffer
@@ -582,12 +917,10 @@ buffer in your bug report.
 
   ;; Dump buffer local variables.
   (insert "\nlocal variables:\n================")
-  (dolist (buffer
-	   (delq nil
-		 (mapcar
-		  (lambda (b)
-                    (when (string-match-p "\\*tramp/" (buffer-name b)) b))
-		  (buffer-list))))
+  (dolist (buffer (tramp-compat-seq-keep
+		   (lambda (b)
+		     (when (string-match-p "\\*tramp/" (buffer-name b)) b))
+		   (buffer-list)))
     (let ((reporter-eval-buffer buffer)
 	  (elbuf (get-buffer-create " *tmp-reporter-buffer*")))
       (with-current-buffer elbuf
@@ -595,26 +928,32 @@ buffer in your bug report.
 	(erase-buffer)
 	(insert (format "\n;; %s\n(setq-local\n" (buffer-name buffer)))
 	(lisp-indent-line)
-	(dolist
-	    (varsym
-	     (sort
-	      (append
-	       (mapcar
-		#'intern
-		(all-completions "tramp-" (buffer-local-variables buffer)))
-	       ;; Non-tramp variables of interest.
-	       '(connection-local-variables-alist default-directory))
-	      #'string<))
-	    (reporter-dump-variable varsym elbuf))
+	(dolist (varsym
+		 (sort
+		  (append
+		   (mapcar
+		    #'intern
+		    (all-completions "tramp-" (buffer-local-variables buffer)))
+		   ;; Non-tramp variables of interest.
+		   '(connection-local-variables-alist default-directory))
+		  #'string<))
+	  (reporter-dump-variable varsym elbuf))
 	(lisp-indent-line)
 	(insert ")\n"))
       (insert-buffer-substring elbuf)))
 
+  ;; Beautify encoded values.
+  (goto-char (point-min))
+  (while (search-forward-regexp
+	  (rx "'" (group "(decode-coding-string")) nil 'noerror)
+    (replace-match "\\1"))
+  (goto-char (point-max))
+
   ;; Dump load-path shadows.
   (insert "\nload-path shadows:\n==================\n")
   (ignore-errors
     (mapc
-     (lambda (x) (when (string-match-p "tramp" x) (insert x "\n")))
+     (lambda (x) (when (string-search "tramp" x) (insert x "\n")))
      (split-string (list-load-path-shadows t) "\n")))
 
   ;; Append buffers only when we are in message mode.
@@ -622,7 +961,7 @@ buffer in your bug report.
 	 (eq major-mode 'message-mode)
 	 (bound-and-true-p mml-mode))
 
-    (let ((tramp-buf-regexp "\\*\\(debug \\)?tramp/")
+    (let ((tramp-buf-regexp (rx "*" (? "debug ") "tramp/"))
 	  (buffer-list (tramp-list-tramp-buffers))
 	  (curbuf (current-buffer)))
 
@@ -633,7 +972,7 @@ buffer in your bug report.
 	(setq buffer-read-only nil)
 	(goto-char (point-min))
 	(while (not (eobp))
-	  (if (re-search-forward tramp-buf-regexp (point-at-eol) t)
+	  (if (search-forward-regexp tramp-buf-regexp (line-end-position) t)
 	      (forward-line 1)
 	    (forward-line 0)
 	    (let ((start (point)))
@@ -659,9 +998,9 @@ the debug buffer(s).")
 	(setq buffer-read-only t)
 	(goto-char (point-min))
 
-	(when (y-or-n-p "Do you want to append the buffer(s)? ")
+	(when (y-or-n-p "Do you want to append the buffer(s)?")
 	  ;; OK, let's send.  First we delete the buffer list.
-	  (kill-buffer nil)
+	  (kill-buffer)
 	  (switch-to-buffer curbuf)
 	  (goto-char (point-max))
 	  (insert (propertize "\n" 'display "\n\
